@@ -2,24 +2,34 @@
 // for details. All rights reserved. Use of this source code is governed by a
 // BSD-style license that can be found in the LICENSE file.
 
-// @dart = 2.6
-
 // part of "core_patch.dart";
 
+// This function takes care of rehashing of the expandos in [objects]. We
+// do this eagerly after snapshot deserialization.
+@pragma("vm:entry-point", "call")
+void _rehashObjects(List objects) {
+  final int length = objects.length;
+  for (int i = 0; i < length; ++i) {
+    objects[i]._rehash();
+  }
+}
+
 @patch
+@pragma("vm:entry-point")
 class Expando<T> {
   @patch
-  Expando([String name])
+  Expando([String? name])
       : name = name,
-        _data = new List(_minSize),
+        _data = new List<_WeakProperty?>.filled(_minSize, null),
         _used = 0;
 
   static const _minSize = 8;
-  static final _deletedEntry = new _WeakProperty(null, null);
+  static final _deletedEntry = new _WeakProperty();
 
   @patch
-  T operator [](Object object) {
-    _checkType(object);
+  T? operator [](Object object) {
+    // TODO(http://dartbug.com/48634): Rename to `key`.
+    checkValidWeakTarget(object, 'object');
 
     var mask = _size - 1;
     var idx = object._identityHashCode & mask;
@@ -27,7 +37,7 @@ class Expando<T> {
 
     while (wp != null) {
       if (identical(wp.key, object)) {
-        return wp.value;
+        return unsafeCast<T?>(wp.value);
       } else if (wp.key == null) {
         // This entry has been cleared by the GC.
         _data[idx] = _deletedEntry;
@@ -40,8 +50,9 @@ class Expando<T> {
   }
 
   @patch
-  void operator []=(Object object, T value) {
-    _checkType(object);
+  void operator []=(Object object, T? value) {
+    // TODO(http://dartbug.com/48634): Rename to `key`.
+    checkValidWeakTarget(object, 'object');
 
     var mask = _size - 1;
     var idx = object._identityHashCode & mask;
@@ -84,7 +95,10 @@ class Expando<T> {
     }
 
     if (_used < _limit) {
-      _data[idx] = new _WeakProperty(object, value);
+      var ephemeron = new _WeakProperty();
+      ephemeron.key = object;
+      ephemeron.value = value;
+      _data[idx] = ephemeron;
       _used++;
       return;
     }
@@ -118,7 +132,7 @@ class Expando<T> {
 
     // Reset the mappings to empty so that we can just add the existing
     // valid entries.
-    _data = new List(new_size);
+    _data = new List<_WeakProperty?>.filled(new_size, null);
     _used = 0;
 
     for (var i = 0; i < old_data.length; i++) {
@@ -135,19 +149,33 @@ class Expando<T> {
     }
   }
 
-  static _checkType(object) {
-    if ((object == null) ||
-        (object is bool) ||
-        (object is num) ||
-        (object is String)) {
-      throw new ArgumentError.value(object,
-          "Expandos are not allowed on strings, numbers, booleans or null");
-    }
-  }
+  int get _size => _data.length;
+  int get _limit => (3 * (_size ~/ 4));
 
-  get _size => _data.length;
-  get _limit => (3 * (_size ~/ 4));
-
-  List _data;
+  List<_WeakProperty?> _data;
   int _used; // Number of used (active and deleted) slots.
 }
+
+@patch
+class WeakReference<T extends Object> {
+  @patch
+  factory WeakReference(T target) = _WeakReferenceImpl<T>;
+}
+
+@pragma("vm:entry-point")
+class _WeakReferenceImpl<T extends Object> implements WeakReference<T> {
+  _WeakReferenceImpl(T target) {
+    checkValidWeakTarget(target, 'target');
+    _target = target;
+  }
+
+  @pragma("vm:recognized", "other")
+  @pragma("vm:prefer-inline")
+  @pragma("vm:external-name", "WeakReference_getTarget")
+  external T? get target;
+
+  @pragma("vm:recognized", "other")
+  @pragma("vm:prefer-inline")
+  @pragma("vm:external-name", "WeakReference_setTarget")
+  external set _target(T? value);
+}
