@@ -2,8 +2,6 @@
 // for details. All rights reserved. Use of this source code is governed by a
 // BSD-style license that can be found in the LICENSE file.
 
-// @dart = 2.6
-
 /// Note: the VM concatenates all patch files into a single patch file. This
 /// file is the first patch in "dart:isolate" which contains all the imports
 /// used by patches of that library. We plan to change this when we have a
@@ -23,7 +21,8 @@ import "dart:typed_data" show ByteBuffer, TypedData, Uint8List;
 @patch
 class ReceivePort {
   @patch
-  factory ReceivePort() => new _ReceivePortImpl();
+  factory ReceivePort([String debugName = '']) =>
+      new _ReceivePortImpl(debugName);
 
   @patch
   factory ReceivePort.fromRawReceivePort(RawReceivePort rawPort) {
@@ -39,9 +38,10 @@ class Capability {
 
 @pragma("vm:entry-point")
 class _CapabilityImpl implements Capability {
-  factory _CapabilityImpl() native "CapabilityImpl_factory";
+  @pragma("vm:external-name", "CapabilityImpl_factory")
+  external factory _CapabilityImpl();
 
-  bool operator ==(var other) {
+  bool operator ==(Object other) {
     return (other is _CapabilityImpl) && _equals(other);
   }
 
@@ -49,8 +49,10 @@ class _CapabilityImpl implements Capability {
     return _get_hashcode();
   }
 
-  _equals(other) native "CapabilityImpl_equals";
-  _get_hashcode() native "CapabilityImpl_get_hashcode";
+  @pragma("vm:external-name", "CapabilityImpl_equals")
+  external bool _equals(Object other);
+  @pragma("vm:external-name", "CapabilityImpl_get_hashcode")
+  external int _get_hashcode();
 }
 
 @patch
@@ -63,18 +65,20 @@ class RawReceivePort {
    * event is received.
    */
   @patch
-  factory RawReceivePort([Function handler]) {
-    _RawReceivePortImpl result = new _RawReceivePortImpl();
+  factory RawReceivePort([Function? handler, String debugName = '']) {
+    _RawReceivePortImpl result = new _RawReceivePortImpl(debugName);
     result.handler = handler;
     return result;
   }
 }
 
 class _ReceivePortImpl extends Stream implements ReceivePort {
-  _ReceivePortImpl() : this.fromRawReceivePort(new RawReceivePort());
+  _ReceivePortImpl([String debugName = ''])
+      : this.fromRawReceivePort(new RawReceivePort(null, debugName));
 
-  _ReceivePortImpl.fromRawReceivePort(this._rawPort) {
-    _controller = new StreamController(onCancel: close, sync: true);
+  _ReceivePortImpl.fromRawReceivePort(this._rawPort)
+      : _controller = new StreamController(sync: true) {
+    _controller.onCancel = close;
     _rawPort.handler = _controller.add;
   }
 
@@ -82,8 +86,8 @@ class _ReceivePortImpl extends Stream implements ReceivePort {
     return _rawPort.sendPort;
   }
 
-  StreamSubscription listen(void onData(var message),
-      {Function onError, void onDone(), bool cancelOnError}) {
+  StreamSubscription listen(void onData(var message)?,
+      {Function? onError, void onDone()?, bool? cancelOnError}) {
     return _controller.stream.listen(onData,
         onError: onError, onDone: onDone, cancelOnError: cancelOnError);
   }
@@ -94,13 +98,13 @@ class _ReceivePortImpl extends Stream implements ReceivePort {
   }
 
   final RawReceivePort _rawPort;
-  StreamController _controller;
+  final StreamController _controller;
 }
 
 typedef void _ImmediateCallback();
 
 /// The callback that has been registered through `scheduleImmediate`.
-_ImmediateCallback _pendingImmediateCallback;
+_ImmediateCallback? _pendingImmediateCallback;
 
 /// The closure that should be used as scheduleImmediateClosure, when the VM
 /// is responsible for the event loop.
@@ -128,11 +132,20 @@ Function _getIsolateScheduleImmediateClosure() {
 
 @pragma("vm:entry-point")
 class _RawReceivePortImpl implements RawReceivePort {
-  factory _RawReceivePortImpl() native "RawReceivePortImpl_factory";
+  factory _RawReceivePortImpl(String debugName) {
+    final port = _RawReceivePortImpl._(debugName);
+    _portMap[port._get_id()] = <String, dynamic>{
+      'port': port,
+    };
+    return port;
+  }
+
+  @pragma("vm:external-name", "RawReceivePortImpl_factory")
+  external factory _RawReceivePortImpl._(String debugName);
 
   close() {
     // Close the port and remove it from the handler map.
-    _handlerMap.remove(this._closeInternal());
+    _portMap.remove(this._closeInternal());
   }
 
   SendPort get sendPort {
@@ -149,46 +162,60 @@ class _RawReceivePortImpl implements RawReceivePort {
   }
 
   /**** Internal implementation details ****/
-  _get_id() native "RawReceivePortImpl_get_id";
-  _get_sendport() native "RawReceivePortImpl_get_sendport";
+  @pragma("vm:external-name", "RawReceivePortImpl_get_id")
+  external int _get_id();
+  @pragma("vm:external-name", "RawReceivePortImpl_get_sendport")
+  external SendPort _get_sendport();
 
   // Called from the VM to retrieve the handler for a message.
   @pragma("vm:entry-point", "call")
   static _lookupHandler(int id) {
-    var result = _handlerMap[id];
+    var result = _portMap[id]?['handler'];
     return result;
   }
 
-  // Called from the VM to dispatch to the handler.
   @pragma("vm:entry-point", "call")
-  static void _handleMessage(Function handler, var message) {
+  static _lookupOpenPorts() {
+    return _portMap.values.map((e) => e['port']).toList();
+  }
+
+  // Called from the VM to retrieve  the handler and handle a message.
+  @pragma("vm:entry-point", "call")
+  static _handleMessage(int id, var message) {
+    final handler = _portMap[id]?['handler'];
+    if (handler == null) {
+      return null;
+    }
     // TODO(floitsch): this relies on the fact that any exception aborts the
     // VM. Once we have non-fatal global exceptions we need to catch errors
     // so that we can run the immediate callbacks.
     handler(message);
     _runPendingImmediateCallback();
+    return handler;
   }
 
   // Call into the VM to close the VM maintained mappings.
-  _closeInternal() native "RawReceivePortImpl_closeInternal";
+  @pragma("vm:external-name", "RawReceivePortImpl_closeInternal")
+  external int _closeInternal();
 
-  void set handler(Function value) {
-    _handlerMap[this._get_id()] = value;
+  // Set this port as active or inactive in the VM. If inactive, this port
+  // will not be considered live even if it hasn't been explicitly closed.
+  // TODO(bkonyi): determine if we want to expose this as an option through
+  // RawReceivePort.
+  @pragma("vm:external-name", "RawReceivePortImpl_setActive")
+  external _setActive(bool active);
+
+  void set handler(Function? value) {
+    final int id = this._get_id();
+    if (!_portMap.containsKey(id)) {
+      _portMap[id] = <String, dynamic>{
+        'port': this,
+      };
+    }
+    _portMap[id]!['handler'] = value;
   }
 
-  // TODO(iposva): Ideally keep this map in the VM.
-  // id to handler mapping.
-  static _initHandlerMap() {
-    // TODO(18511): Workaround bad CheckSmi hoisting.
-    var tempMap = new HashMap();
-    // Collect feedback that not all keys are Smis.
-    tempMap["."] = 1;
-    tempMap["."] = 2;
-
-    return new HashMap();
-  }
-
-  static final Map _handlerMap = _initHandlerMap();
+  static final _portMap = <int, Map<String, dynamic>>{};
 }
 
 @pragma("vm:entry-point")
@@ -212,32 +239,27 @@ class _SendPortImpl implements SendPort {
   }
 
   /*--- private implementation ---*/
-  _get_id() native "SendPortImpl_get_id";
-  _get_hashcode() native "SendPortImpl_get_hashcode";
+  @pragma("vm:external-name", "SendPortImpl_get_id")
+  external _get_id();
+  @pragma("vm:external-name", "SendPortImpl_get_hashcode")
+  external _get_hashcode();
 
   // Forward the implementation of sending messages to the VM.
-  void _sendInternal(var message) native "SendPortImpl_sendInternal_";
+  @pragma("vm:external-name", "SendPortImpl_sendInternal_")
+  external void _sendInternal(var message);
 }
 
 typedef _NullaryFunction();
-typedef _UnaryFunction(Null args);
-typedef _BinaryFunction(Null args, Null message);
+typedef _UnaryFunction(Never args);
+typedef _BinaryFunction(Never args, Never message);
 
 /**
- * Takes the real entry point as argument and invokes it with the
- * initial message.  Defers execution of the entry point until the
- * isolate is in the message loop.
+ * Takes the real entry point as argument and schedules it to run in the message
+ * queue.
  */
 @pragma("vm:entry-point", "call")
-void _startMainIsolate(Function entryPoint, List<String> args) {
-  _startIsolate(
-      null, // no parent port
-      entryPoint,
-      args,
-      null, // no message
-      true, // isSpawnUri
-      null, // no control port
-      null); // no capabilities
+void _startMainIsolate(Function entryPoint, List<String>? args) {
+  _delayEntrypointInvocation(entryPoint, args, null, true);
 }
 
 /**
@@ -252,47 +274,21 @@ Function _getStartMainIsolateFunction() {
 }
 
 /**
- * Takes the real entry point as argument and invokes it with the initial
- * message.
+ * Takes the real entry point as argument and schedules it to run in the message
+ * queue.
  */
 @pragma("vm:entry-point", "call")
 void _startIsolate(
-    SendPort parentPort,
-    Function entryPoint,
-    List<String> args,
-    var message,
-    bool isSpawnUri,
-    RawReceivePort controlPort,
-    List capabilities) {
-  // The control port (aka the main isolate port) does not handle any messages.
-  if (controlPort != null) {
-    controlPort.handler = (_) {}; // Nobody home on the control port.
-  }
+    Function entryPoint, List<String>? args, Object? message, bool isSpawnUri) {
+  _delayEntrypointInvocation(entryPoint, args, message, isSpawnUri);
+}
 
-  if (parentPort != null) {
-    // Build a message to our parent isolate providing access to the
-    // current isolate's control port and capabilities.
-    //
-    // TODO(floitsch): Send an error message if we can't find the entry point.
-    var readyMessage = new List(2);
-    readyMessage[0] = controlPort.sendPort;
-    readyMessage[1] = capabilities;
-
-    // Out of an excess of paranoia we clear the capabilities from the
-    // stack.  Not really necessary.
-    capabilities = null;
-    parentPort.send(readyMessage);
-  }
-  assert(capabilities == null);
-
-  // Delay all user code handling to the next run of the message loop. This
-  // allows us to intercept certain conditions in the event dispatch, such as
-  // starting in paused state.
-  RawReceivePort port = new RawReceivePort();
+void _delayEntrypointInvocation(Function entryPoint, List<String>? args,
+    Object? message, bool allowZeroOneOrTwoArgs) {
+  final port = RawReceivePort();
   port.handler = (_) {
     port.close();
-
-    if (isSpawnUri) {
+    if (allowZeroOneOrTwoArgs) {
       if (entryPoint is _BinaryFunction) {
         (entryPoint as dynamic)(args, message);
       } else if (entryPoint is _UnaryFunction) {
@@ -304,7 +300,6 @@ void _startIsolate(
       entryPoint(message);
     }
   };
-  // Make sure the message handler is triggered.
   port.sendPort.send(null);
 }
 
@@ -317,15 +312,15 @@ class Isolate {
   static Isolate get current => _currentIsolate;
 
   @patch
-  String get debugName => _getDebugName(controlPort);
+  String? get debugName => _getDebugName(controlPort);
 
   @patch
-  static Future<Uri> get packageRoot {
+  static Future<Uri?> get packageRoot {
     return Future.value(null);
   }
 
   @patch
-  static Future<Uri> get packageConfig {
+  static Future<Uri?> get packageConfig {
     var hook = VMLibraryHooks.packageConfigUriFuture;
     if (hook == null) {
       throw new UnsupportedError("Isolate.packageConfig");
@@ -334,7 +329,7 @@ class Isolate {
   }
 
   @patch
-  static Future<Uri> resolvePackageUri(Uri packageUri) {
+  static Future<Uri?> resolvePackageUri(Uri packageUri) {
     var hook = VMLibraryHooks.resolvePackageUriFuture;
     if (hook == null) {
       throw new UnsupportedError("Isolate.resolvePackageUri");
@@ -343,75 +338,78 @@ class Isolate {
   }
 
   static bool _packageSupported() =>
-      (VMLibraryHooks.packageRootUriFuture != null) &&
       (VMLibraryHooks.packageConfigUriFuture != null) &&
       (VMLibraryHooks.resolvePackageUriFuture != null);
 
   @patch
   static Future<Isolate> spawn<T>(void entryPoint(T message), T message,
-      {bool paused: false,
-      bool errorsAreFatal,
-      SendPort onExit,
-      SendPort onError,
-      String debugName}) async {
+      {bool paused = false,
+      bool errorsAreFatal = true,
+      SendPort? onExit,
+      SendPort? onError,
+      String? debugName}) async {
     // `paused` isn't handled yet.
-    RawReceivePort readyPort;
-    try {
-      // Check for the type of `entryPoint` on the spawning isolate to make
-      // error-handling easier.
-      if (entryPoint is! _UnaryFunction) {
-        throw new ArgumentError(entryPoint);
-      }
-      // The VM will invoke [_startIsolate] with entryPoint as argument.
-      readyPort = new RawReceivePort();
+    // Check for the type of `entryPoint` on the spawning isolate to make
+    // error-handling easier.
+    if (entryPoint is! _UnaryFunction) {
+      throw new ArgumentError(entryPoint);
+    }
+    // The VM will invoke [_startIsolate] with entryPoint as argument.
 
-      // We do not inherit the package config settings from the parent isolate,
-      // instead we use the values that were set on the command line.
-      var packageConfig = VMLibraryHooks.packageConfigString;
-      var script = VMLibraryHooks.platformScript;
-      if (script == null) {
-        // We do not have enough information to support spawning the new
-        // isolate.
-        throw new UnsupportedError("Isolate.spawn");
-      }
-      if (script.scheme == "package") {
+    // We do not inherit the package config settings from the parent isolate,
+    // instead we use the values that were set on the command line.
+    var packageConfig = VMLibraryHooks.packageConfigString;
+    var script = VMLibraryHooks.platformScript;
+    if (script == null) {
+      // We do not have enough information to support spawning the new
+      // isolate.
+      throw new UnsupportedError("Isolate.spawn");
+    }
+    if (script.isScheme("package")) {
+      if (Isolate._packageSupported()) {
+        // resolving script uri is not really neccessary, but can be useful
+        // for better failed-to-lookup-function-in-a-script spawn errors.
         script = await Isolate.resolvePackageUri(script);
       }
+    }
 
-      _spawnFunction(
-          readyPort.sendPort,
-          script.toString(),
-          entryPoint,
-          message,
-          paused,
-          errorsAreFatal,
-          onExit,
-          onError,
-          null,
-          packageConfig,
-          debugName);
+    final RawReceivePort readyPort =
+        new RawReceivePort(null, 'Isolate.spawn ready');
+    try {
+      _spawnFunction(readyPort.sendPort, script.toString(), entryPoint, message,
+          paused, errorsAreFatal, onExit, onError, packageConfig, debugName);
       return await _spawnCommon(readyPort);
     } catch (e, st) {
-      if (readyPort != null) {
-        readyPort.close();
-      }
+      readyPort.close();
       return await new Future<Isolate>.error(e, st);
     }
   }
 
+  @pragma("vm:external-name", "Isolate_spawnFunction")
+  external static void _spawnFunction(
+      SendPort readyPort,
+      String uri,
+      Function topLevelFunction,
+      var message,
+      bool paused,
+      bool errorsAreFatal,
+      SendPort? onExit,
+      SendPort? onError,
+      String? packageConfig,
+      String? debugName);
+
   @patch
   static Future<Isolate> spawnUri(Uri uri, List<String> args, var message,
-      {bool paused: false,
-      SendPort onExit,
-      SendPort onError,
-      bool errorsAreFatal,
-      bool checked,
-      Map<String, String> environment,
-      Uri packageRoot,
-      Uri packageConfig,
-      bool automaticPackageResolution: false,
-      String debugName}) async {
-    RawReceivePort readyPort;
+      {bool paused = false,
+      SendPort? onExit,
+      SendPort? onError,
+      bool errorsAreFatal = true,
+      bool? checked,
+      Map<String, String>? environment,
+      Uri? packageRoot,
+      Uri? packageConfig,
+      bool automaticPackageResolution = false,
+      String? debugName}) async {
     if (environment != null) {
       throw new UnimplementedError("environment");
     }
@@ -434,38 +432,31 @@ class Isolate {
             "packageRoot and a packageConfig.");
       }
     }
+    // Resolve the uri against the current isolate's root Uri first.
+    final Uri spawnedUri = _rootUri!.resolveUri(uri);
+
+    // Inherit this isolate's package resolution setup if not overridden.
+    if (!automaticPackageResolution && packageConfig == null) {
+      if (Isolate._packageSupported()) {
+        packageConfig = await Isolate.packageConfig;
+      }
+    }
+
+    // Ensure to resolve package: URIs being handed in as parameters.
+    if (packageConfig != null) {
+      // Avoid calling resolvePackageUri if not strictly necessary in case
+      // the API is not supported.
+      if (packageConfig.isScheme("package")) {
+        packageConfig = await Isolate.resolvePackageUri(packageConfig);
+      }
+    }
+
+    // The VM will invoke [_startIsolate] and not `main`.
+    final packageConfigString = packageConfig?.toString();
+
+    final RawReceivePort readyPort =
+        new RawReceivePort(null, 'Isolate.spawnUri ready');
     try {
-      // Resolve the uri against the current isolate's root Uri first.
-      var spawnedUri = _rootUri.resolveUri(uri);
-
-      // Inherit this isolate's package resolution setup if not overridden.
-      if (!automaticPackageResolution &&
-          (packageRoot == null) &&
-          (packageConfig == null)) {
-        if (Isolate._packageSupported()) {
-          packageRoot = await Isolate.packageRoot;
-          packageConfig = await Isolate.packageConfig;
-        }
-      }
-
-      // Ensure to resolve package: URIs being handed in as parameters.
-      if (packageRoot != null) {
-        // `packages/` directory is no longer supported. Force it null.
-        // TODO(mfairhurst) Should this throw an exception?
-        packageRoot = null;
-      } else if (packageConfig != null) {
-        // Avoid calling resolvePackageUri if not strictly necessary in case
-        // the API is not supported.
-        if (packageConfig.scheme == "package") {
-          packageConfig = await Isolate.resolvePackageUri(packageConfig);
-        }
-      }
-
-      // The VM will invoke [_startIsolate] and not `main`.
-      readyPort = new RawReceivePort();
-      var packageRootString = packageRoot?.toString();
-      var packageConfigString = packageConfig?.toString();
-
       _spawnUri(
           readyPort.sendPort,
           spawnedUri.toString(),
@@ -478,20 +469,17 @@ class Isolate {
           checked,
           null,
           /* environment */
-          packageRootString,
           packageConfigString,
           debugName);
       return await _spawnCommon(readyPort);
     } catch (e) {
-      if (readyPort != null) {
-        readyPort.close();
-      }
+      readyPort.close();
       rethrow;
     }
   }
 
   static Future<Isolate> _spawnCommon(RawReceivePort readyPort) {
-    Completer completer = new Completer<Isolate>.sync();
+    final completer = new Completer<Isolate>.sync();
     readyPort.handler = (readyMessage) {
       readyPort.close();
       if (readyMessage is List && readyMessage.length == 2) {
@@ -527,42 +515,34 @@ class Isolate {
   static const _DEL_ERROR = 8;
   static const _ERROR_FATAL = 9;
 
-  static void _spawnFunction(
-      SendPort readyPort,
-      String uri,
-      Function topLevelFunction,
-      var message,
-      bool paused,
-      bool errorsAreFatal,
-      SendPort onExit,
-      SendPort onError,
-      String packageRoot,
-      String packageConfig,
-      String debugName) native "Isolate_spawnFunction";
+  // For 'spawnFunction' see internal_patch.dart.
 
-  static void _spawnUri(
+  @pragma("vm:external-name", "Isolate_spawnUri")
+  external static void _spawnUri(
       SendPort readyPort,
       String uri,
       List<String> args,
       var message,
       bool paused,
-      SendPort onExit,
-      SendPort onError,
+      SendPort? onExit,
+      SendPort? onError,
       bool errorsAreFatal,
-      bool checked,
-      List environment,
-      String packageRoot,
-      String packageConfig,
-      String debugName) native "Isolate_spawnUri";
+      bool? checked,
+      List? environment,
+      String? packageConfig,
+      String? debugName);
 
-  static void _sendOOB(port, msg) native "Isolate_sendOOB";
+  @pragma("vm:external-name", "Isolate_sendOOB")
+  external static void _sendOOB(port, msg);
 
-  static String _getDebugName(SendPort controlPort)
-      native "Isolate_getDebugName";
+  @pragma("vm:external-name", "Isolate_getDebugName")
+  external static String _getDebugName(SendPort controlPort);
 
   @patch
   void _pause(Capability resumeCapability) {
-    var msg = new List(4)
+    // _sendOOB expects a fixed length array and hence we create a fixed
+    // length array and assign values to it instead of using [ ... ].
+    var msg = new List<Object?>.filled(4, null)
       ..[0] = 0 // Make room for OOB message type.
       ..[1] = _PAUSE
       ..[2] = pauseCapability
@@ -572,7 +552,7 @@ class Isolate {
 
   @patch
   void resume(Capability resumeCapability) {
-    var msg = new List(4)
+    var msg = new List<Object?>.filled(4, null)
       ..[0] = 0 // Make room for OOB message type.
       ..[1] = _RESUME
       ..[2] = pauseCapability
@@ -581,8 +561,8 @@ class Isolate {
   }
 
   @patch
-  void addOnExitListener(SendPort responsePort, {Object response}) {
-    var msg = new List(4)
+  void addOnExitListener(SendPort responsePort, {Object? response}) {
+    var msg = new List<Object?>.filled(4, null)
       ..[0] = 0 // Make room for OOB message type.
       ..[1] = _ADD_EXIT
       ..[2] = responsePort
@@ -592,7 +572,7 @@ class Isolate {
 
   @patch
   void removeOnExitListener(SendPort responsePort) {
-    var msg = new List(3)
+    var msg = new List<Object?>.filled(3, null)
       ..[0] = 0 // Make room for OOB message type.
       ..[1] = _DEL_EXIT
       ..[2] = responsePort;
@@ -601,7 +581,7 @@ class Isolate {
 
   @patch
   void setErrorsFatal(bool errorsAreFatal) {
-    var msg = new List(4)
+    var msg = new List<Object?>.filled(4, null)
       ..[0] = 0 // Make room for OOB message type.
       ..[1] = _ERROR_FATAL
       ..[2] = terminateCapability
@@ -611,7 +591,7 @@ class Isolate {
 
   @patch
   void kill({int priority: beforeNextEvent}) {
-    var msg = new List(4)
+    var msg = new List<Object?>.filled(4, null)
       ..[0] = 0 // Make room for OOB message type.
       ..[1] = _KILL
       ..[2] = terminateCapability
@@ -620,8 +600,9 @@ class Isolate {
   }
 
   @patch
-  void ping(SendPort responsePort, {Object response, int priority: immediate}) {
-    var msg = new List(5)
+  void ping(SendPort responsePort,
+      {Object? response, int priority: immediate}) {
+    var msg = new List<Object?>.filled(5, null)
       ..[0] = 0 // Make room for OOM message type.
       ..[1] = _PING
       ..[2] = responsePort
@@ -632,7 +613,7 @@ class Isolate {
 
   @patch
   void addErrorListener(SendPort port) {
-    var msg = new List(3)
+    var msg = new List<Object?>.filled(3, null)
       ..[0] = 0 // Make room for OOB message type.
       ..[1] = _ADD_ERROR
       ..[2] = port;
@@ -641,7 +622,7 @@ class Isolate {
 
   @patch
   void removeErrorListener(SendPort port) {
-    var msg = new List(3)
+    var msg = new List<Object?>.filled(3, null)
       ..[0] = 0 // Make room for OOB message type.
       ..[1] = _DEL_ERROR
       ..[2] = port;
@@ -655,10 +636,10 @@ class Isolate {
         terminateCapability: portAndCapabilities[2]);
   }
 
-  static List _getPortAndCapabilitiesOfCurrentIsolate()
-      native "Isolate_getPortAndCapabilitiesOfCurrentIsolate";
+  @pragma("vm:external-name", "Isolate_getPortAndCapabilitiesOfCurrentIsolate")
+  external static List _getPortAndCapabilitiesOfCurrentIsolate();
 
-  static Uri _getCurrentRootUri() {
+  static Uri? _getCurrentRootUri() {
     try {
       return Uri.parse(_getCurrentRootUriStr());
     } catch (e) {
@@ -666,7 +647,40 @@ class Isolate {
     }
   }
 
-  static String _getCurrentRootUriStr() native "Isolate_getCurrentRootUriStr";
+  @pragma("vm:external-name", "Isolate_getCurrentRootUriStr")
+  external static String _getCurrentRootUriStr();
+
+  @pragma("vm:external-name", "Isolate_exit_")
+  external static Never _exit(SendPort? finalMessagePort, Object? message);
+
+  static Never exit([SendPort? finalMessagePort, Object? message]) {
+    _exit(finalMessagePort, message);
+  }
+
+  /**
+   * Creates an Uri representing the script which was compiled into kernel
+   * binary in [kernelBlob].
+   * The resulting Uri can be used for the subsequent spawnUri calls.
+   * Such spawnUri will start an isolate which would run the given
+   * compiled script in [kernelBlob].
+   */
+  /*static*/ Uri createUriForKernelBlob(Uint8List kernelBlob) {
+    return Uri.parse(_registerKernelBlob(kernelBlob));
+  }
+
+  /**
+   * Unregisters kernel blob previously registered with
+   * [createUriForKernelBlob] and frees underlying resources.
+   */
+  /*static*/ void unregisterKernelBlobUri(Uri kernelBlobUri) {
+    _unregisterKernelBlob(kernelBlobUri.toString());
+  }
+
+  @pragma("vm:external-name", "Isolate_registerKernelBlob")
+  external static String _registerKernelBlob(Uint8List kernelBlob);
+
+  @pragma("vm:external-name", "Isolate_unregisterKernelBlob")
+  external static void _unregisterKernelBlob(String kernelBlobUri);
 }
 
 @patch
@@ -688,13 +702,13 @@ abstract class TransferableTypedData {
 
 @pragma("vm:entry-point")
 class _TransferableTypedDataImpl implements TransferableTypedData {
-  factory _TransferableTypedDataImpl(List<TypedData> list)
-      native "TransferableTypedData_factory";
+  @pragma("vm:external-name", "TransferableTypedData_factory")
+  external factory _TransferableTypedDataImpl(List<TypedData> list);
 
   ByteBuffer materialize() {
     return _materializeIntoUint8List().buffer;
   }
 
-  Uint8List _materializeIntoUint8List()
-      native "TransferableTypedData_materialize";
+  @pragma("vm:external-name", "TransferableTypedData_materialize")
+  external Uint8List _materializeIntoUint8List();
 }
